from __future__ import print_function
import keras
from keras.layers import Dense, Conv2D, BatchNormalization, Activation
from keras.layers import AveragePooling2D, Input, Flatten
from keras.optimizers import Adam
from keras.callbacks import ModelCheckpoint, LearningRateScheduler
from keras.callbacks import ReduceLROnPlateau
from keras.preprocessing.image import ImageDataGenerator
from keras.regularizers import l2
from keras import backend as K
from keras.models import Model
from keras.datasets import cifar10
import numpy as np
import os
import time
import argparse
import sys
from keras.models import load_model
import time
import tensorflow as tf
from tropic_util import *
from os import listdir
from os.path import isfile, join
from sklearn.model_selection import train_test_split
from sklearn import metrics
#from sklearn.cross_validation import train_test_split
from tflearn.data_utils import image_preloader
#from tflearn.data_utils import directory_to_samples
#import tflearn.data_utils

#np.set_printoptions(threshold=np.nan)


def write_to_logfile(logfile_report, dataset_name,exp_name,epochs,mse, rmse, mae , mae_count, mae_missed_classify, n_diff, total_time):

    import os
    cwd = os.getcwd()
    print(cwd)
    print("_____write to logfile_report : %s"%(logfile_report))
    
    with open(logfile_report,'a') as logfile:
        logfile.write('%s, %s, epochs: %s, mse: %.3f, rmse: %.3f, mae: %.3f, mae_count: %.3f, mae_missed_class: %.2f, wrong_image: %i, train_time: %.2f' %(dataset_name, exp_name, epochs, mse, rmse, mae, mae_count, mae_missed_classify, n_diff, total_time))
        logfile.write("\n")

def workon_fruitamount_prediction(y_test, y_pred):
    y_test_numfruit = np.zeros(16)
    y_pred_numfruit = np.zeros(16)
    n_correct_prediction = 0
    
    n = len(y_test) #300 or num of testing images
    for i in range(n):
        actual_nfruit = int(y_test[i])
        y_test_numfruit[actual_nfruit] +=1
        pred_nfruit = int(np.around(y_pred[i]))
        if (actual_nfruit == pred_nfruit):
            y_pred_numfruit[actual_nfruit] +=1
            n_correct_prediction +=1
            print ('image : %i, with #of fruit %i, correct prediction' %(i, actual_nfruit))
        else:
            print ('image : %i, with #of fruit %i, predicted as % i ==> wrong prediction' %(i, actual_nfruit, pred_nfruit))
    
    for i in range(1,16):
        print(i, '  ' , int(y_test_numfruit[i]), '   ', int(y_pred_numfruit[i]))
    print('\n number images of correct prediction #fruit:', n_correct_prediction)
        
        
    
    
def custom_loss(y_test_raw, y_pred_raw):
    
    n = len(y_test_raw)
    num_classes = len(y_test_raw[0])

    #loss for mse_count __________________________
    print('\n\n______1. [INFO] loss for mae_count (only consider for the correct class and count number of fruit)')
    y_test = np.amax(y_test_raw, axis=1)
    y_test_class = y_test_raw.argmax(axis=1)

    print('\nfruit class ')
    print(y_test_class)

    print('\nfruit amount (y_test)')
    print(y_test)

    y_pred = np.zeros(n)
    for i in range(n):
        y_pred[i] = y_pred_raw[i , y_test_class[i]]

    print('\nfruit amount (y_pred)')
    print(np.round(y_pred))
    
    workon_fruitamount_prediction(y_test, y_pred) #for bar_graph report
    
    mae_count = metrics.mean_absolute_error(y_test, y_pred)
    print('mae_count :', mae_count)

    #missed classification loss__________________________
    print('\n\n______2. [INFO] missed classification loss (only consider the missed classes loss)_______')
    y_test = np.zeros(shape=(n,num_classes-1))
    #print('y_test')
    #print(y_test)

    y_pred =np.zeros(shape=(n,num_classes-1))
    for i in range(n):
        y_temp = y_pred_raw[i]
        y_temp = np.delete(y_temp, y_test_class[i])
        y_pred[i] = y_temp
    
    print('y_pred, missed classify')
    print(y_pred)
    mae_missed_classify = metrics.mean_absolute_error(y_test, y_pred)
    print('\n mae_missed_classify :', mae_missed_classify)
    
    
    #check if missed -classify, based on y_pred_class (maximum number of fruit)
    print('\n\n_____3. [INFO] predicted fruit class, based on maximum #fruit predicted')
    
    print('y_test_class')
    print(y_test_class)
    print('\ny_pred_class, based on maximum #fruit predicted')
    y_pred_class = y_pred_raw.argmax(axis=1)
    print(y_pred_class)

    diff = np.isclose(y_test_class, y_pred_class)
    n_diff = n - np.sum(diff)
    print('\ndifferent between y_test_class and y_pred_class')
    print(diff)
    print('\n there are %i/%i images incorrected prediction base on maximum of fruit\n' %(n_diff,n))


    
    return(mae_count, mae_missed_classify, n_diff)

                   
def evaluate_model(savedmodel_dir, model, x_test, y_test, count_only):

    if model==None: # must manual provide model here
        
        os.chdir(savedmodel_dir)    #move to saved model dir
        
        sortmodel = sorted(filter(os.path.isfile, os.listdir('.')), key=os.path.getmtime)
        for i,model_name in enumerate(reversed(sortmodel)):
            print ('loading model: %s'% (model_name))
            model = load_model(str(model_name))
            break
        os.chdir("../..")   #move back to working directory
        
    
    print('model :',model)
    y_pred=model.predict(x_test)
    #y_predround = np.around(y_pred, decimals =1)
    y_predround = np.around(y_pred)

    print('\n ________ y_test and y_pred _______________')    
    yi = 0
    for yt in y_test:
        print('image: ', yi, ' ' , yt,'\t', y_predround[yi])
        yi = yi+1
        
    if count_only: #only care about number of fruit, no consider class
        mse = metrics.mean_squared_error(y_test, y_pred)
        print('mse for count_only: ', mse)
        
        return (0, mse, 0, 0)
                
    else:   #main idea program, regression etc ..
        
        
        mse = metrics.mean_squared_error(y_test, y_pred)
        rmse = np.sqrt(metrics.mean_squared_error(y_test, y_pred))
        mae = metrics.mean_absolute_error(y_test, y_pred)
        
        (mae_count, mae_missed_classify, n_diff)   = custom_loss(y_test, y_pred)
                
        print('saved model dir : ', savedmodel_dir)
        print('_________ loss value ________')
        
        print('MSE : ', mse)
        print('RMSE : ', rmse)
        print('MAE : ', mae)
        print('MAE_count: ', mae_count)
        print('MAE_missed_classify: ', mae_missed_classify)
        print('n_diff (#of image): ',n_diff)
        
        return (mse, rmse, mae, mae_count, mae_missed_classify, n_diff)


########################


def str2bool(v):
    return v.lower() in ("yes","true")

            
def random_shuffle(images, labels):
    perm = np.arange(len(images))
    np.random.shuffle(perm)
    random_image = images[perm]
    random_label = labels[perm]
    return random_image, random_label


def get_fruit_count(work_dir, images, labels):
#return label as number of fruit (for count only)
    i = 0
    
    countfruit = []
    for im in images:
        s2 = im.rsplit('/',1)[1]
        s3 = s2.replace('jpg','txt')
        s4 = work_dir+s3
        count = len(open(s4).readlines(  ))        
        countfruit.append(count)
    
    return countfruit


def get_label_count(work_dir, images, labels):
#return label as n-hot
    i = 0
    
    print('work_dir : ',work_dir)
    for im in images:
        s2 = im.rsplit('/',1)[1]
        s3 = s2.replace('jpg','txt')
        #s4 = work_dir+'count/'+s3
        s4 = work_dir+s3
        count = len(open(s4).readlines(  ))
        #print(s4,' ' ,count)
        
        l = labels[i]
        l[l==1] = count
        i = i+1

    return labels


def load_dataset(dataset_name, exp_name, imw, is_train, count_only=False, is_onehot=False):
    
    np.set_printoptions(threshold=sys.maxsize)
    
    work_dir='./dataset/'+dataset_name+"/"+exp_name+"/"

    train_dir = work_dir+'train_images'
    test_dir = work_dir+'test_images'

    print("train_dir: ", train_dir)
    print("test_dir: ", test_dir)
    
    
    x_train, y_train, train_images = image_preloader(train_dir, image_shape=(imw, imw),  grayscale=False,  mode='folder', categorical_labels=True, normalize=True)
    x_test, y_test, test_images = image_preloader(test_dir, image_shape=(imw, imw), grayscale=False, mode='folder', categorical_labels=True,  normalize=True)  

    
    print('_________train_images : ', len(train_images))
    #print(train_images, len(train_images) )       
    
    print('_________test_images : ', len(test_images))
    #print(test_images, len(test_images) )       
    
    x_train = np.asarray(x_train[:])
    y_train = np.asarray(y_train[:])
    x_test = np.asarray(x_test[:])
    y_test = np.asarray(y_test[:])
    
    #up till now xtrain, xtest is one-hot label, if for counting to this modify label
    if is_onehot==False:    # for counting, or count-and-classify
        if(count_only): #only count fruit in images, no consider class, return just a number of fruit
            #y_train = np.max(y_train, axis=1)
            #y_test = np.max(y_test, axis=1)
            y_train = get_fruit_count(work_dir+'train_txt/', train_images,y_train)
            y_test = get_fruit_count(work_dir+'test_txt/', test_images,y_test)
        else: #count and classify (n-hot label)
            y_train = get_label_count(work_dir+'train_txt/', train_images,y_train)
            y_test = get_label_count(work_dir+'test_txt/', test_images,y_test)
        
    for i in range(len(train_images)):
        print('image: ', i,' ' , y_train[i],  train_images[i])
    
    print('\n___________testing images___________\n')
    for i in range(len(test_images)):
        print('image: ', i,' ' ,y_test[i],  test_images[i])

    
    #if is_train: 
    print('[INFO] subtract pixel mean ....')
    #subtract  pixel mean
    x_train_mean = np.mean(x_train, axis=0)
    #print('____[INFO] x_train_mean_____: ', (x_train_mean))
    x_train -= x_train_mean
    x_test -= x_train_mean
    

    return x_train,y_train, x_test, y_test
    

'''    
######## evaluate model for classification only
def evaluate_classification_model(savedmodel_dir, model, x_test, y_test,x_testdemo,y_testdemo):

    if model==None: # must manual provide model here

        os.chdir(savedmodel_dir)
        sortmodel = sorted(filter(os.path.isfile, os.listdir('.')), key=os.path.getmtime)
        for i,model_name in enumerate(reversed(sortmodel)):
            print ('loading model: %s'% (model_name))
            model = load_model(str(model_name))
            break
        
    print('model :',model)
    print('\n  _______________result from y_test, y_pred and dataset_demo')
    y_pred=model.predict(x_test)
    y_predround = np.around(y_pred, decimals =1)

    #dataset_demo
    y_preddemo=model.predict(x_testdemo)
    y_predrounddemo = np.around(y_preddemo, decimals =1)
    
    
    print('\n ________ y_test and y_pred ')    
    yi = 0
    for yt in y_test:
        print(yt,'\t', y_predround[yi])
        yi = yi+1
    
    print('\n ________ dataset demo ')    
    ya = 0
    for yt in y_testdemo:
        print(yt,'\t', y_predrounddemo[ya])
        ya  = ya+1    
    
    print('\n ________ y_test _______________')
    print(np.argmax(y_test,1))
    print('\n ________ y_pred _______________')
    print(np.argmax(y_pred,1))
    
    print('\n ________ y_test from dataset_demo _______________')
    print(np.argmax(y_testdemo,1))    
    print('\n ________ y_pred from dataset_demo _______________')
    print(np.argmax(y_preddemo,1))    
    
    #print to compare predicted result for count and classify

    ytest= np.argmax(y_test,1)
    ypred = np.argmax(y_pred,1)
    yn = len(ytest)
    correct = np.sum(ytest==ypred)
    acc = (correct/yn)*100.0
    print('Test accuracy : %f' %(acc))
 

    ytestdemo= np.argmax(y_testdemo,1)
    ypreddemo = np.argmax(y_preddemo,1)
    yn = len(ytestdemo)
    correct = np.sum(ytestdemo==ypreddemo)
    acc = (correct/yn)*100.0
    print('Test accuracy (from dataset_demo) : %f' %(acc))
    
       
    return acc

'''


'''
def load_dataset_fortesting(dataset_name, exp_name, imw, is_train, count_only=False, is_onehot=False):
    
    np.set_printoptions(threshold=sys.maxsize)
    #np.set_printoptions(np.nan)
    work_dir='./dataset/'+dataset_name+"/"

    train_dir = work_dir+'train'
    test_dir = work_dir+'test'
    test_dirdemo = '/data/2019/counting/dataset/dataset_demo/test'

    print("train_dir: ", train_dir)
    print("test_dir: ", test_dir)
    

    x_train, y_train, train_images = image_preloader(train_dir, image_shape=(imw, imw),  grayscale=False,  mode='folder', categorical_labels=True, normalize=True)
    x_test, y_test, test_images = image_preloader(test_dir, image_shape=(imw, imw), grayscale=False, mode='folder', categorical_labels=True,  normalize=True)  
    x_testdemo, y_testdemo, test_imagesdemo = image_preloader(test_dirdemo, image_shape=(imw, imw), grayscale=False, mode='folder', categorical_labels=True,  normalize=True)  

    
    print('_________train_images : ', len(train_images))
    #print(train_images, len(train_images) )       
    
    print('_________test_images : ', len(test_images))
    #print(test_images, len(test_images) )       
    
    x_train = np.asarray(x_train[:])
    y_train = np.asarray(y_train[:])
    x_test = np.asarray(x_test[:])
    y_test = np.asarray(y_test[:])
    x_testdemo = np.asarray(x_testdemo[:])
    y_testdemo = np.asarray(y_testdemo[:])
    
    #up till now xtrain, xtest is one-hot label, if for counting to this modify label
    if is_onehot==False:    # for counting, or count-and-classify
        if(count_only):
            #y_train = np.max(y_train, axis=1)
            #y_test = np.max(y_test, axis=1)
            y_train = get_fruit_count(work_dir+'count_train/', train_images,y_train)
            y_test = get_fruit_count(work_dir+'count_test/', test_images,y_test)
        else: #count and classify (n-hot label)
            y_train = get_label_count(work_dir+'count_train/', train_images,y_train)
            y_test = get_label_count(work_dir+'count_test/', test_images,y_test)
        
    for i in range(len(train_images)):
        print(y_train[i],  train_images[i])
    for i in range(len(test_images)):
        print(y_test[i],  test_images[i])

    
    #if is_train: 
    print('[INFO] subtract pixel mean ....')
    #subtract  pixel mean
    x_train_mean = np.mean(x_train, axis=0)
    x_train -= x_train_mean
    x_test -= x_train_mean
    x_testdemo -=x_train_mean
    

    return x_train,y_train, x_test, y_test,x_testdemo, y_testdemo
'''
